""" Toplevel Subgrounds module

This module implements the toplevel API that most developers will be using when
querying The Graph with Subgrounds.
"""

import logging
import warnings
from functools import cached_property
from json import JSONDecodeError
from typing import Any, Type

import httpx
import pandas as pd
from pipe import map, traverse

from ..dataframe_utils import df_of_json
from ..errors import GraphQLError, ServerError
from ..pagination import LegacyStrategy, PaginationStrategy
from ..query import DataRequest, DataResponse, DocumentResponse
from ..subgraph import FieldPath, Subgraph
from ..utils import default_header
from .base import SubgroundsBase

logger = logging.getLogger("subgrounds")
warnings.simplefilter("default")

HTTP2_SUPPORT = True


class AsyncSubgrounds(SubgroundsBase):
    @cached_property
    def _client(self):
        """Cached client"""

        return httpx.AsyncClient(http2=HTTP2_SUPPORT, timeout=self.timeout)

    async def load(self, url: str, save_schema: bool = False, is_subgraph: bool = True):
        """Performs introspection on the provided GraphQL API ``url`` to get the
        schema, stores the schema if ``save_schema`` is ``True`` and returns a
        generated class representing the GraphQL endpoint with all its entities.
        """

        try:
            loader = self._load(url, save_schema, is_subgraph)
            url, query = next(loader)  # if this fails, schema is loaded from cache
            data = await self._fetch(url, {"query": query})
            loader.send(data)

        except StopIteration as e:
            return e.value

        assert False

    async def load_subgraph(self, url: str, save_schema: bool = False) -> Subgraph:
        """Performs introspection on the provided GraphQL API ``url`` to get the
        schema, stores the schema if ``save_schema`` is ``True`` and returns a
        generated class representing the subgraph with all its entities.

        Args:
          url: The url of the API
          save_schema: Flag indicating whether or not the schema should be cached to
            disk. Defaults to False.
          cache_dir: If ``save_schema == True``, then subgraph schemas will be stored
            under ``cache_dir``. Defaults to ``schemas/``

        Returns:
          Subgraph: A generated class representing the subgraph and its entities
        """

        return await self.load(url, save_schema, True)

    async def load_api(self, url: str, save_schema: bool = False) -> Subgraph:
        """Performs introspection on the provided GraphQL API ``url`` to get the
         schema, stores the schema if ``save_schema`` is ``True`` and returns a
         generated class representing the GraphQL endpoint with all its entities.

        Args:
          url: The url of the API
          save_schema: Flag indicating whether or not the schema should be saved
           to disk. Defaults to False.

        Returns:
          A generated class representing the subgraph and its entities
        """

        return await self.load(url, save_schema, False)

    async def execute(
        self,
        req: DataRequest,
        pagination_strategy: Type[PaginationStrategy] | None = LegacyStrategy,
    ) -> DataResponse:
        """Executes a :class:`DataRequest` and returns a :class:`DataResponse`.

        Args:
          req: The :class:`DataRequest` object to be executed
          pagination_strategy: A Class implementing the :class:`PaginationStrategy`
            ``Protocol``. If ``None``, then automatic pagination is disabled.
            Defaults to :class:`LegacyStrategy`.

        Returns:
          A :class:`DataResponse` object representing the response
        """

        try:
            executor = self._execute(req, pagination_strategy)

            doc = next(executor)
            while True:
                data = await self._fetch(
                    doc.url, {"query": doc.graphql, "variables": doc.variables}
                )
                doc = executor.send(DocumentResponse(url=doc.url, data=data))

        except StopIteration as e:
            return e.value

    async def query_json(
        self,
        fpaths: FieldPath | list[FieldPath],
        pagination_strategy: Type[PaginationStrategy] | None = LegacyStrategy,
    ) -> list[dict[str, Any]]:
        """See :func:`~subgrounds.Subgrounds.query_json`.

        Args:
          fpaths: One or more :class:`FieldPath` objects that should be included in
            the request.
          pagination_strategy: A Class implementing the :class:`PaginationStrategy`
            ``Protocol``. If ``None``, then automatic pagination is disabled.
            Defaults to :class:`LegacyStrategy`.

        Returns:
          The reponse data
        """

        fpaths = list([fpaths] | traverse | map(FieldPath._auto_select) | traverse)
        req = self.mk_request(fpaths)
        data = await self.execute(req, pagination_strategy)

        return [doc.data for doc in data.responses]

    async def query_df(
        self,
        fpaths: FieldPath | list[FieldPath],
        columns: list[str] | None = None,
        concat: bool = False,
        pagination_strategy: Type[PaginationStrategy] | None = LegacyStrategy,
    ) -> pd.DataFrame | list[pd.DataFrame]:
        """See :func:`~subgrounds.Subgrounds.query_df`.

        Args:
          fpaths: One or more `FieldPath` objects that should be included in the request
          columns: The column labels.
          merge: Whether or not to merge resulting dataframes.
          pagination_strategy: A class implementing the :class:`PaginationStrategy`
            ``Protocol``. If ``None``, then automatic pagination is disabled.

        Returns:
          A :class:`pandas.DataFrame` containing the reponse data.
        """

        fpaths = list([fpaths] | traverse | map(FieldPath._auto_select) | traverse)
        json_data = await self.query_json(fpaths, pagination_strategy)

        return df_of_json(json_data, fpaths, columns, concat)

    async def query(
        self,
        fpaths: FieldPath | list[FieldPath],
        unwrap: bool = True,
        pagination_strategy: Type[PaginationStrategy] | None = LegacyStrategy,
    ) -> str | int | float | bool | list | tuple | None:
        """See :func:`~subgrounds.Subgrounds.query`.

        Args:
          fpaths: One or more ``FieldPath`` object(s) to query.
          unwrap: Flag indicating whether or not, in the case where the returned data
            is a list of one element, the element itself should be returned instead of
            the list. Defaults to ``True``.
          pagination_strategy: A Class implementing the :class:`PaginationStrategy`
            ``Protocol``. If ``None``, then automatic pagination is disabled.
            Defaults to :class:`LegacyStrategy`.

        Returns:
          The ``FieldPath`` object(s) data
        """

        fpaths = list([fpaths] | traverse | map(FieldPath._auto_select) | traverse)
        blob = await self.query_json(fpaths, pagination_strategy=pagination_strategy)

        def f(fpath: FieldPath) -> dict[str, Any]:
            data = fpath._extract_data(blob)

            if type(data) == list and len(data) == 1 and unwrap:
                return data[0]

            return data

        data = tuple(fpaths | map(f))

        if len(data) == 1:
            return data[0]

        return data

    async def _fetch(self, url: str, blob: dict[str, Any]) -> dict[str, Any]:
        resp = await self._client.post(
            url, json=blob, headers=default_header(url) | self.headers
        )
        resp.raise_for_status()

        try:
            raw_data = resp.json()

        except JSONDecodeError:
            raise ServerError(
                f"Server ({url}) did not respond with proper JSON"
                f"\nDid you query a proper GraphQL endpoint?"
                f"\n\n{resp.content}"
            )

        if (data := raw_data.get("data")) is None:
            raise GraphQLError(raw_data.get("errors", "Unknown Error(s) Found"))

        return data

    async def __aenter__(self):
        await self._client.__aenter__()
        return self

    async def __aexit__(self, *args):
        await self._client.__aexit__(*args)



from abc import ABC
from typing import Any

import plotly.graph_objects as go
from plotly.basedatatypes import BaseTraceType

from subgrounds import FieldPath


class TraceWrapper(ABC):
    graph_object: BaseTraceType

    fpaths: dict[str, FieldPath]
    args: dict[str, Any]

    def __init__(self, **kwargs) -> None:
        self.fpaths = {}
        self.args = {}

        for key, arg in kwargs.items():
            match arg:
                case FieldPath():
                    self.fpaths[key] = arg
                case _:
                    self.args[key] = arg

    def mk_trace(self, data: list[dict[str, Any]] | dict[str, Any]) -> BaseTraceType:
        fpath_data = {}
        for key, fpath in self.fpaths.items():
            item = fpath._extract_data(data)

            if type(item) == list and len(item) == 1:
                fpath_data[key] = item[0]
            else:
                fpath_data[key] = item

        return self.graph_object(**(fpath_data | self.args))  # type: ignore

    @property
    def field_paths(self) -> list[FieldPath]:
        return [fpath for _, fpath in self.fpaths.items()]


# Simple
class Scatter(TraceWrapper):
    """See https://plotly.com/python/line-and-scatter/"""

    graph_object = go.Scatter  # type: ignore


class Pie(TraceWrapper):
    """See https://plotly.com/python/pie-charts/"""

    graph_object = go.Pie  # type: ignore


class Bar(TraceWrapper):
    """See https://plotly.com/python/bar-charts/"""

    graph_object = go.Bar  # type: ignore


class Heatmap(TraceWrapper):
    """See https://plotly.com/python/heatmaps/"""

    graph_object = go.Heatmap  # type: ignore


class Contour(TraceWrapper):
    """See https://plotly.com/python/contour-plots/"""

    graph_object = go.Contour  # type: ignore


class Table(TraceWrapper):
    """See https://plotly.com/python/contour-plots/"""

    graph_object = go.Table  # type: ignore


# Distributions
class Box(TraceWrapper):
    """See https://plotly.com/python/box-plots/"""

    graph_object = go.Box  # type: ignore


class Violin(TraceWrapper):
    """See https://plotly.com/python/violin/"""

    graph_object = go.Violin  # type: ignore


class Histogram(TraceWrapper):
    """See https://plotly.com/python/histograms/"""

    graph_object = go.Histogram  # type: ignore


class Histogram2d(TraceWrapper):
    """See https://plotly.com/python/2D-Histogram/"""

    graph_object = go.Histogram2d  # type: ignore


class Histogram2dContour(TraceWrapper):
    """See https://plotly.com/python/2d-histogram-contour/"""

    graph_object = go.Histogram2dContour  # type: ignore


# Finance
class Ohlc(TraceWrapper):
    """See https://plotly.com/python/ohlc-charts/"""

    graph_object = go.Ohlc  # type: ignore


class Candlestick(TraceWrapper):
    """See https://plotly.com/python/candlestick-charts/"""

    graph_object = go.Candlestick  # type: ignore


class Waterfall(TraceWrapper):
    """See https://plotly.com/python/waterfall-charts/"""

    graph_object = go.Waterfall  # type: ignore


class Funnel(TraceWrapper):
    """See https://plotly.com/python/funnel-charts/"""

    graph_object = go.Funnel  # type: ignore


class Indicator(TraceWrapper):
    """See https://plotly.com/python/indicator/"""

    graph_object = go.Indicator  # type: ignore


# 3d
class Scatter3d(TraceWrapper):
    """See https://plotly.com/python/3d-scatter-plots/"""

    graph_object = go.Scatter3d  # type: ignore


class Surface(TraceWrapper):
    """See https://plotly.com/python/3d-surface-plots/"""

    graph_object = go.Surface  # type: ignore


# Maps
class Scattergeo(TraceWrapper):
    """See https://plotly.com/python/scatter-plots-on-maps/"""

    graph_object = go.Scattergeo  # type: ignore


class Choropleth(TraceWrapper):
    """See https://plotly.com/python/choropleth-maps/"""

    graph_object = go.Choropleth  # type: ignore


class Scattermapbox(TraceWrapper):
    """See https://plotly.com/python/scattermapbox/"""

    graph_object = go.Scattermapbox  # type: ignore


class Choroplethmapbox(TraceWrapper):
    """See https://plotly.com/python/mapbox-county-choropleth/"""

    graph_object = go.Choroplethmapbox  # type: ignore


class Densitymapbox(TraceWrapper):
    """See https://plotly.com/python/mapbox-density-heatmaps/"""

    graph_object = go.Densitymapbox  # type: ignore


# Specialized
class Scatterpolar(TraceWrapper):
    """See https://plotly.com/python/polar-chart/"""

    graph_object = go.Scatterpolar  # type: ignore


class Barpolar(TraceWrapper):
    """See https://plotly.com/python/wind-rose-charts/"""

    graph_object = go.Barpolar  # type: ignore


class Sunburst(TraceWrapper):
    """See https://plotly.com/python/sunburst-charts/"""

    graph_object = go.Sunburst  # type: ignore


class Treemap(TraceWrapper):
    """See https://plotly.com/python/treemaps/"""

    graph_object = go.Treemap  # type: ignore


class Icicle(TraceWrapper):
    """See https://plotly.com/python/icicle-charts/"""

    graph_object = go.Icicle  # type: ignore


class Sankey(TraceWrapper):
    """See https://plotly.com/python/sankey-diagram/"""

    graph_object = go.Sankey  # type: ignore


class Parcoords(TraceWrapper):
    """See https://plotly.com/python/parallel-coordinates-plot/"""

    graph_object = go.Parcoords  # type: ignore


class Parcats(TraceWrapper):
    """See https://plotly.com/python/parallel-categories-diagram/"""

    graph_object = go.Parcats  # type: ignore


class Carpet(TraceWrapper):
    """See https://plotly.com/python/carpet-plot/"""

    graph_object = go.Carpet  # type: ignore


class Scattercarpet(TraceWrapper):
    """See https://plotly.com/python/carpet-scatter/"""

    graph_object = go.Scattercarpet  # type: ignore


class Contourcarpet(TraceWrapper):
    """See https://plotly.com/python/carpet-contour/"""

    graph_object = go.Contourcarpet  # type: ignore



from __future__ import annotations

from typing import Any, ClassVar

import pandas as pd
from dash import dash_table, dcc, html
from dash.dependencies import Input, Output
from pipe import map, where

from subgrounds import FieldPath, Subgrounds
from subgrounds.contrib.plotly import Figure

from .abcs import Refreshable


class Graph(dcc.Graph, Refreshable):
    counter: ClassVar[int] = 0
    wrapped_figure: Figure

    def __init__(self, fig: Figure, **kwargs) -> None:
        super().__init__(id=f"graph-{Graph.counter}", figure=fig.figure, **kwargs)
        Graph.counter += 1
        self.wrapped_figure = fig

    @property
    def dash_dependencies(self) -> list[Output]:
        return [Output(self.id, "figure")]

    @property
    def dash_dependencies_outputs(self) -> list[Any]:
        self.wrapped_figure.refresh()
        return [self.wrapped_figure.figure]


class DataTable(dash_table.DataTable, Refreshable):
    counter: ClassVar[int] = 0

    subgrounds: Subgrounds
    data: list[FieldPath]
    columns: list[str] | None
    concat: bool
    append: bool
    df: pd.DataFrame | list[pd.DataFrame] | None

    def __init__(
        self,
        subgrounds: Subgrounds,
        data: FieldPath | list[FieldPath],
        columns: list[str] | None = None,
        concat: bool = False,
        append: bool = False,
        **kwargs,
    ):
        self.subgrounds = subgrounds
        self.fpaths = data if type(data) == list else [data]
        self.column_names = columns
        self.concat = concat
        self.append = append
        self.df = None

        super().__init__(id=f"datatable-{DataTable.counter}", **kwargs)
        DataTable.counter += 1

        self.refresh()

    def refresh(self) -> None:
        match (self.df, self.append):
            case (None, _) | (_, False):
                self.df = self.subgrounds.query_df(
                    self.fpaths, columns=self.column_names, concat=self.concat
                )
            case (_, True):
                self.df = pd.concat(
                    [
                        self.df,
                        self.subgrounds.query_df(
                            self.fpaths, columns=self.column_names, concat=self.concat
                        ),
                    ],
                    ignore_index=True,
                )
                self.df = self.df.drop_duplicates()

        self.columns = [{"name": i, "id": i} for i in self.df.columns]
        self.data = self.df.to_dict("records")

    @property
    def dash_dependencies(self) -> list[Output]:
        return [Output(self.id, "data")]

    @property
    def dash_dependencies_outputs(self) -> list[Output]:
        self.refresh()
        return [self.df.to_dict("records")]


class AutoUpdate(html.Div):
    counter: ClassVar[int] = 0

    def __init__(self, app, sec_interval: int = 1, children=[], **kwargs):
        id = f"interval-{AutoUpdate.counter}"

        super().__init__(
            children=[
                dcc.Interval(id=id, interval=sec_interval * 1000, n_intervals=0),
                *children,
            ],
            **kwargs,
        )

        AutoUpdate.counter += 1

        def flatten(input: list[Any]):
            return [item for sublist in input for item in sublist]

        subgrounds_children = list(
            children | where(lambda child: isinstance(child, Refreshable))
        )
        deps = flatten(
            list(subgrounds_children | map(lambda child: child.dash_dependencies))
        )

        def update(n):
            outputs = flatten(
                list(
                    subgrounds_children
                    | map(lambda child: child.dash_dependencies_outputs)
                )
            )

            return outputs[0] if len(outputs) == 1 else outputs

        # Register callback
        app.callback(*deps, Input(id, "n_intervals"))(update)



""" Pandas DataFrame utility module containing functions related to the
formatting of GraphQL JSON data into DataFrames.
"""

from __future__ import annotations

from dataclasses import dataclass
from functools import partial
from itertools import cycle
from typing import Any

import pandas as pd
from pipe import dedup, groupby, map, traverse, where
from typing_extensions import Self

from subgrounds.query import Selection
from subgrounds.subgraph import FieldPath
from subgrounds.utils import union


def gen_columns(data: list | dict, prefix: str = "") -> list[str]:
    match data:
        case dict():
            return list(
                list(data.keys())
                | map(
                    lambda key: gen_columns(
                        data[key], f"{prefix}_{key}" if prefix != "" else key
                    )
                )
                | traverse
            )
        case list():
            return gen_columns(data[0], prefix)
        case _:
            return prefix


def fmt_cols(df: pd.DataFrame, col_map: dict[str, str]) -> pd.DataFrame:
    df = df.rename(col_map, axis="columns")
    cols = list(col_map.values() | dedup | where(lambda name: name in df.columns))
    return df[cols]


@dataclass
class DataFrameColumns:
    """Helper class that holds data related to the shape of a DataFrame"""

    key: str
    fpaths: list[str]

    def combine(self, other: Self) -> Self:
        """Returns new DataFrameColumns containing the union of :attr:`self` and
        :attr:`other`'s columns

        Args:
          other: Columns to be combined to :attr:`self`

        Returns:
          New :class:`Self` containing the union of :attr:`self` and :attr:`other`
        """
        return DataFrameColumns(self.key, union(self.fpaths, other.fpaths))

    def mk_df(
        self, data: list[dict[str, Any]], path_map: dict[str, FieldPath]
    ) -> pd.DataFrame:
        """Formats the JSON data :attr:`data` into a DataFrame containing the columns
        defined in :attr:`self`.

        Args:
          data: The JSON data to be formatted into a dataframe
          path_map: A dictionary of :attr:`(key-FieldPath)` pairs

        Returns:
          The JSON data formatted into a DataFrame
        """
        cols_data = {
            col: path_map[col]._extract_data(data)
            for col in self.fpaths
            if col in path_map
        }

        rows_data = []

        def mk_rows(data: dict, row: dict = {}):
            if all([type(d) != list for d in list(data.values())]):
                rows_data.append(data | row)
            else:
                non_list_items = {
                    key: value for key, value in data.items() if type(value) != list
                }
                list_items = {
                    key: value for key, value in data.items() if type(value) == list
                }
                length = len(list(list_items.values())[0])
                for i in range(length):
                    mk_rows(
                        data={key: value[i] for key, value in list_items.items()},
                        row=row | non_list_items,
                    )

        mk_rows(cols_data, row={})
        return pd.DataFrame(data=rows_data)


def columns_of_selections(selections: list[Selection]) -> list[DataFrameColumns]:
    """Generates a list of DataFrame columns specifications based on a list of
    :class:`Selection` trees.

    Args:
      selections: The selection trees

    Returns:
      The list of DataFrame columns specifications
    """

    def columns_of_selections(
        selections: list[Selection], keys: list[str] = [], fpaths: list[str] = []
    ) -> list[DataFrameColumns]:
        if len(selections) > 0:
            non_list_selections = [
                select for select in selections if not select.contains_list()
            ]
            non_list_fpaths = list(
                non_list_selections
                | map(
                    lambda select: [
                        "_".join([*keys, *path]) for path in select.data_paths
                    ]
                )
                | traverse
            )
        else:
            non_list_fpaths = ["_".join(keys)]

        list_selections = [select for select in selections if select.contains_list()]

        if list_selections == []:
            return [DataFrameColumns("_".join(keys), fpaths + non_list_fpaths)]
        else:
            return list(
                list_selections
                | map(
                    lambda select: columns_of_selections(
                        select.selection,
                        keys=[*keys, select.key],
                        fpaths=fpaths + non_list_fpaths,
                    )
                )
            )

    return list(columns_of_selections(selections) | traverse)


def df_of_json(
    json_data: list[dict[str, Any]],
    fpaths: list[FieldPath],
    columns: list[str] | None = None,
    concat: bool = False,
) -> pd.DataFrame | list[pd.DataFrame]:
    """Formats the JSON data :attr:`json_data` into Pandas DataFrames,
    flattening the data in the process.

    Depending on the request's fieldpaths, either one or multiple dataframes will
    be returned based on how flattenable the response data is.

    :attr:`fpaths` is a list of :class:`FieldPath` objects corresponding to the
    set of fieldpaths that were used to get the response data :attr:`json_data`.

    :attr:`columns` is an optional argument used to rename the dataframes(s)
    columns. The length of :attr:`columns` must be the same as the number of columns
    of all returned dataframes.

    :attr:`concat` indicates whether or not the resulting dataframes should be
    concatenated together. The dataframes must have the same number of columns,
    as well as the same column names (which can be set using the :attr:`columns`
    argument).

    Args:
      json_data: Response data
      fpaths: Fieldpaths that yielded the response data
      columns: Column names. Defaults to None.
      concat: Flag indicating whether or not to concatenate the resulting dataframes,
        if there are more than one. Defaults to False.

    Returns:
      The resulting dataframe(s)
    """

    if columns is None:
        columns = list(fpaths | map(lambda fpath: fpath._name()))

    col_fpaths = zip(fpaths, cycle(columns))
    col_map = {fpath._name(use_aliases=True): colname for fpath, colname in col_fpaths}

    path_map = {fpath._name(use_aliases=True): fpath for fpath in fpaths}

    dfs = list(
        fpaths
        | groupby(lambda fpath: fpath._subgraph._url)
        | map(lambda group: FieldPath._merge(group[1]))
        | map(columns_of_selections)
        | traverse
        | map(partial(DataFrameColumns.mk_df, data=json_data, path_map=path_map))
    )

    if len(dfs) == 0:
        return pd.DataFrame(columns=columns, data=[])

    if len(dfs) == 1:
        return fmt_cols(dfs[0], col_map)

    if concat:
        return pd.concat(
            list(dfs | map(lambda df: fmt_cols(df, col_map))),
            ignore_index=True,
        )

    return list(dfs | map(lambda df: fmt_cols(df, col_map)))



from .base import DocumentTransform, RequestTransform
from .apply import apply_transforms
from .defaults import DEFAULT_GLOBAL_TRANSFORMS, DEFAULT_SUBGRAPH_TRANSFORMS
from .transforms import LocalSyntheticField, TypeTransform

__all__ = [
    "apply_transforms",
    "DocumentTransform",
    "LocalSyntheticField",
    "RequestTransform",
    "TypeTransform",
    "DEFAULT_GLOBAL_TRANSFORMS",
    "DEFAULT_SUBGRAPH_TRANSFORMS",
]



from __future__ import annotations

import logging
from typing import Generator, cast

from pipe import chain_with, map

from subgrounds.query import DataRequest, DataResponse

from .base import DocumentTransform, RequestTransform
from .transforms import DocumentRequestTransform

logger = logging.getLogger("subgrounds")
TransformGen = Generator[DataRequest | DataResponse, DataRequest | DataResponse, None]


def handle_transform(
    transform: RequestTransform,
) -> Generator[None | DataRequest | DataResponse, DataRequest | DataResponse, None]:
    """This function bundles the transform request and response as a generator.

    The follow is as follows:
    -> :class:`subgrounds.query.DataRequest` is sent in
    -> It is transformed, and then sent out
    -> Then, in an infinite loop:
      -> :class:`subgrounds.query.DataResponse` is sent in
      -> It is transformed, then sent out

    The infinite loop allows us to continously transform responses,
      - Needed by `execute_iter`
    """

    req = cast(DataRequest, (yield))
    yield transform.transform_request(req)

    while True:
        resp = cast(DataResponse, (yield))
        yield transform.transform_response(req, resp)


def apply_transforms(
    request_transforms: list[RequestTransform],
    document_transforms: dict[str, list[DocumentTransform]],
    req: DataRequest,
) -> Generator[None | DataRequest | DataResponse, DataResponse, None]:
    """Apply all `RequestTransforms` and `DocumentTransforms` to a `DataRequest` and a
     corresponding `DataResponse`.

    This function abstractly applies a series of transforms onto a request and response.
    The execution of the request is handled outside this function (ala. sans-io), which
     allows this function to only work with the abstracted components.

    Note: For simplification, all `DocumentTransforms` stored at the subgraph are
     converted to specific `RequestTransforms` when applied.
    """

    unique_doc_urls = {doc.url for doc in req.documents}

    # Iterating through all unique document urls, get each document's transforms from
    #  the subgraph converting all `DocumentTransforms` into `DocumentRequestTransforms`
    # We do this to make it easier for us to only work with one type of transform.
    converted_transforms = (
        DocumentRequestTransform(transform, url)
        for url in unique_doc_urls
        for transform in document_transforms[url]
    )

    # Construct our list of generators from our two sets of transforms.
    # The request transforms are before the (converted) document transforms.
    stack: list[TransformGen] = list(
        request_transforms | chain_with(converted_transforms) | map(handle_transform)
    )

    # Go through every transform in the stack and transform the request
    for transform in stack:
        next(transform)  # ditch `None`
        req = cast(DataRequest, transform.send(req))

    # yield the final transformed request (valid "graphql")
    yield req

    # We enter an infinite loop here to allow multiple documents to be transformed
    #  back up the stack since `execute_iter` produces them page-by-page to be streamed.
    while True:
        # retrieve the response (from the `executor` governing this generator)
        resp = yield

        # Finally, using the response, iterate through the transforms in reverse order,
        #  transforming the raw response up back through the transforms
        for transform in reversed(stack):
            next(transform)  # ditch `None`
            resp = cast(DataResponse, transform.send(resp))

        # Take the final transformed response and send it back to the `executor`
        yield resp



""" Toplevel Subgrounds module

This module implements the toplevel API that most developers will be using when
querying The Graph with Subgrounds.
"""

import functools
import logging
import warnings
from collections.abc import Iterator
from contextlib import suppress
from functools import cached_property
from json import JSONDecodeError
from typing import Any, Type, cast

import httpx
import pandas as pd
from pipe import map, traverse

from ..dataframe_utils import df_of_json
from ..errors import GraphQLError, ServerError
from ..pagination import LegacyStrategy, PaginationStrategy
from ..query import DataRequest, DataResponse, Document, DocumentResponse
from ..subgraph import FieldPath, Subgraph
from ..utils import default_header
from .base import SubgroundsBase

logger = logging.getLogger("subgrounds")
warnings.simplefilter("default")

HTTP2_SUPPORT = True


class Subgrounds(SubgroundsBase):
    @cached_property
    def _client(self):
        """Cached client"""

        return httpx.Client(http2=HTTP2_SUPPORT, timeout=self.timeout)

    def load(
        self,
        url: str,
        save_schema: bool = False,
        cache_dir: str | None = None,
        is_subgraph: bool = True,
    ) -> Subgraph:
        if cache_dir is not None:
            warnings.warn("This will be depreciated", DeprecationWarning)

        try:
            loader = self._load(url, save_schema, is_subgraph)
            url, query = next(loader)  # if this fails, schema is loaded from cache
            data = self._fetch(url, {"query": query})
            loader.send(data)

        except StopIteration as e:
            return e.value

        assert False

    def load_subgraph(
        self, url: str, save_schema: bool = False, cache_dir: str | None = None
    ) -> Subgraph:
        """Performs introspection on the provided GraphQL API ``url`` to get the
        schema, stores the schema if ``save_schema`` is ``True`` and returns a
        generated class representing the subgraph with all its entities.

        Args:
          url The url of the API.
          save_schema: Flag indicating whether or not the schema should be cached to
            disk.

        Returns:
          Subgraph: A generated class representing the subgraph and its entities
        """

        return self.load(url, save_schema, cache_dir, True)

    def load_api(
        self, url: str, save_schema: bool = False, cache_dir: str | None = None
    ) -> Subgraph:
        """Performs introspection on the provided GraphQL API ``url`` to get the
         schema, stores the schema if ``save_schema`` is ``True`` and returns a
         generated class representing the GraphQL endpoint with all its entities.

        Args:
          url: The url of the API.
          save_schema: Flag indicating whether or not the schema should be saved
           to disk.

        Returns:
          A generated class representing the subgraph and its entities
        """

        return self.load(url, save_schema, cache_dir, False)

    def execute(
        self,
        req: DataRequest,
        pagination_strategy: Type[PaginationStrategy] | None = LegacyStrategy,
    ) -> DataResponse:
        """Executes a :class:`DataRequest` and returns a :class:`DataResponse`.

        Args:
          req: The :class:`DataRequest` object to be executed.
          pagination_strategy: A Class implementing the :class:`PaginationStrategy`
            ``Protocol``. If ``None``, then automatic pagination is disabled.
            Defaults to :class:`LegacyStrategy`.

        Returns:
          A :class:`DataResponse` object representing the response
        """

        try:
            executor = self._execute(req, pagination_strategy)

            doc = next(executor)
            while True:
                data = self._fetch(
                    doc.url, {"query": doc.graphql, "variables": doc.variables}
                )
                doc = executor.send(DocumentResponse(url=doc.url, data=data))

        except StopIteration as e:
            return e.value

    def execute_iter(
        self,
        req: DataRequest,
        pagination_strategy: Type[PaginationStrategy] | None = LegacyStrategy,
    ) -> Iterator[DocumentResponse]:
        """Same as `execute`, except that an iterator is returned which will iterate
        the data pages.

        Args:
          req: The :class:`DataRequest` object to be executed
          pagination_strategy: A Class implementing the :class:`PaginationStrategy`
           ``Protocol``. If ``None``, then automatic pagination is disabled.
           Defaults to :class:`LegacyStrategy`.

        Returns:
          An iterator over the :class:`DocumentResponse` pages.

        ⚠️ DOES NOT apply global transforms across multiple documents or their pages.
         Since we yield each page as we get it, it's not possible to accurately perform
         the transforms since we don't collect the pages. This means transforms
         expecting multiple documents or pages of documents will be inaccurate.
        """

        with suppress(StopIteration):
            executor = self._execute_iter(req, pagination_strategy)

            while True:
                doc = cast(Document, next(executor))
                data = self._fetch(
                    doc.url, {"query": doc.graphql, "variables": doc.variables}
                )

                yield cast(
                    DocumentResponse,
                    executor.send(DocumentResponse(url=doc.url, data=data)),
                )

    def query_json(
        self,
        fpaths: FieldPath | list[FieldPath],
        pagination_strategy: Type[PaginationStrategy] | None = LegacyStrategy,
    ) -> list[dict[str, Any]]:
        """Equivalent to
         ``Subgrounds.execute(Subgrounds.mk_request(fpaths), pagination_strategy)``.

        Args:
          fpaths: One or more :class:`FieldPath` objects
            that should be included in the request.
          pagination_strategy: A Class implementing the :class:`PaginationStrategy`
            ``Protocol``. If ``None``, then automatic pagination is disabled.
            Defaults to :class:`LegacyStrategy`.

        Returns:
          The reponse data
        """

        fpaths = list([fpaths] | traverse | map(FieldPath._auto_select) | traverse)
        req = self.mk_request(fpaths)
        data = self.execute(req, pagination_strategy)
        return [doc.data for doc in data.responses]

    def query_json_iter(
        self,
        fpaths: FieldPath | list[FieldPath],
        pagination_strategy: Type[PaginationStrategy] | None = LegacyStrategy,
    ) -> Iterator[dict[str, Any]]:
        """Same as `query_json` returns an iterator over the response data pages.

        Args:
          fpaths: One or more :class:`FieldPath` objects that should be included
            in the request.
          pagination_strategy: A Class implementing the :class:`PaginationStrategy`
            ``Protocol``. If ``None``, then automatic pagination is disabled.
            Defaults to :class:`LegacyStrategy`.

        Returns:
          list[dict[str, Any]]: The reponse data
        """

        fpaths = list([fpaths] | traverse | map(FieldPath._auto_select) | traverse)
        req = self.mk_request(fpaths)
        responses = self.execute_iter(req, pagination_strategy)

        for resp in responses:
            yield resp.data

    def query_df(
        self,
        fpaths: FieldPath | list[FieldPath],
        columns: list[str] | None = None,
        concat: bool = False,
        pagination_strategy: Type[PaginationStrategy] | None = LegacyStrategy,
    ) -> pd.DataFrame | list[pd.DataFrame]:
        """Same as :func:`Subgrounds.query` but formats the response data into a
        Pandas DataFrame. If the response data cannot be flattened to a single query
        (e.g.: when querying multiple list fields that return different entities),
        then multiple dataframes are returned

        ``fpaths`` is a list of :class:`FieldPath` objects that indicate which
        data must be queried.

        ``columns`` is an optional argument used to rename the dataframes(s)
        columns. The length of ``columns`` must be the same as the number of columns
        of *all* returned dataframes.

        ``concat`` indicates whether or not the resulting dataframes should be
        concatenated together. The dataframes must have the same number of columns,
        as well as the same column names and types (the names can be set using the
        ``columns`` argument).

        Args:
          fpaths: One or more `FieldPath` objects that should be included in the request
          columns: The column labels. Defaults to None.
          merge: Whether or not to merge resulting dataframes.
          pagination_strategy: A Class implementing the :class:`PaginationStrategy`
            ``Protocol``. If ``None``, then automatic pagination is disabled.
            Defaults to :class:`LegacyStrategy`.

        Returns:
          A :class:`pandas.DataFrame` containing the reponse data.

        Example:

        .. code-block:: python

            >>> from subgrounds import Subgrounds
            >>> sg = Subgrounds()
            >>> univ3 = sg.load_subgraph(
            ...    'https://api.thegraph.com/subgraphs/name/uniswap/uniswap-v3')

            # Define price SyntheticField
            >>> univ3.Swap.price = abs(univ3.Swap.amount0) / abs(univ3.Swap.amount1)

            # Query last 10 swaps from the ETH/USDC pool
            >>> eth_usdc = univ3.Query.swaps(
            ...     orderBy=univ3.Swap.timestamp,
            ...     orderDirection='desc',
            ...     first=10,
            ...     where=[
            ...         univ3.Swap.pool == '0x8ad599c3a0ff1de082011efddc58f1908eb6e6d8'
            ...     ]
            ... )
            >>> sg.query_df([
            ...     eth_usdc.timestamp,
            ...     eth_usdc.price
            ... ])
              swaps_timestamp  swaps_price
            0       1643213811  2618.886394
            1       1643213792  2618.814281
            2       1643213792  2617.500494
            3       1643213763  2615.458495
            4       1643213763  2615.876574
            5       1643213739  2615.352390
            6       1643213678  2615.205713
            7       1643213370  2614.115746
            8       1643213210  2613.077301
            9       1643213196  2610.686563
        """

        fpaths = list([fpaths] | traverse | map(FieldPath._auto_select) | traverse)
        json_data = self.query_json(fpaths, pagination_strategy=pagination_strategy)

        return df_of_json(json_data, fpaths, columns, concat)

    def query_df_iter(
        self,
        fpaths: FieldPath | list[FieldPath],
        pagination_strategy: Type[PaginationStrategy] | None = LegacyStrategy,
    ) -> Iterator[pd.DataFrame | list[pd.DataFrame]]:
        """Same as `query_df` except returns an iterator over the response data pages

        Args:
          fpaths: One or more `FieldPath` objects that should be included in the request
          columns: The column labels. Defaults to None.
          merge: Whether or not to merge resulting dataframes.
          pagination_strategy: A Class implementing the :class:`PaginationStrategy`
            ``Protocol``. If ``None``, then automatic pagination is disabled.
            Defaults to :class:`LegacyStrategy`.

        Returns:
          An iterator over the response data pages, each as a :class:`pandas.DataFrame`.
        """

        fpaths = list([fpaths] | traverse | map(FieldPath._auto_select) | traverse)
        for page in self.query_json_iter(fpaths, pagination_strategy):
            yield df_of_json(page, fpaths, None, False)

    def query(
        self,
        fpaths: FieldPath | list[FieldPath],
        unwrap: bool = True,
        pagination_strategy: Type[PaginationStrategy] | None = LegacyStrategy,
    ) -> str | int | float | bool | list | tuple | None:
        """Executes one or multiple ``FieldPath`` objects immediately and returns the
        data (as a tuple if multiple ``FieldPath`` objects are provided).

        Args:
          fpaths: One or more ``FieldPath`` object(s) to query.
          unwrap: Flag indicating whether or not, in the case where the returned data
            is a list of one element, the element itself should be returned instead of
            the list. Defaults to ``True``.
          pagination_strategy: A Class implementing the :class:`PaginationStrategy`
            ``Protocol``. If ``None``, then automatic pagination is disabled.
            Defaults to :class:`LegacyStrategy`.

        Returns:
          The ``FieldPath`` object(s) data

        Example:

        .. code-block:: python

          >>> from subgrounds import Subgrounds
          >>> sg = Subgrounds()
          >>> univ3 = sg.load_subgraph(
          ...  'https://api.thegraph.com/subgraphs/name/uniswap/uniswap-v3')

          # Define price SyntheticField
          >>> univ3.Swap.price = abs(univ3.Swap.amount0) / abs(univ3.Swap.amount1)

          # Construct FieldPath to get price of last swap on ETH/USDC pool
          >>> eth_usdc_last = univ3.Query.swaps(
          ...     orderBy=univ3.Swap.timestamp,
          ...     orderDirection='desc',
          ...     first=1,
          ...     where=[
          ...         univ3.Swap.pool == '0x8ad599c3a0ff1de082011efddc58f1908eb6e6d8'
          ...     ]
          ... ).price

          # Query last price FieldPath
          >>> sg.query(eth_usdc_last)
          2628.975030015892
        """

        fpaths = list([fpaths] | traverse | map(FieldPath._auto_select) | traverse)
        blob = self.query_json(fpaths, pagination_strategy=pagination_strategy)

        def f(fpath: FieldPath) -> dict[str, Any]:
            data = fpath._extract_data(blob)

            if type(data) == list and len(data) == 1 and unwrap:
                return data[0]

            return data

        data = tuple(fpaths | map(f))

        if len(data) == 1:
            return data[0]

        return data

    def query_iter(
        self,
        fpaths: FieldPath | list[FieldPath],
        unwrap: bool = True,
        pagination_strategy: Type[PaginationStrategy] | None = LegacyStrategy,
    ) -> Iterator[str | int | float | bool | list[Any] | tuple | None]:
        """Same as `query` except an iterator over the resonse data pages is returned.

        Args:
          fpath: One or more ``FieldPath`` object(s) to query.
          unwrap: Flag indicating whether or not, in the case where
            the returned data is a list of one element, the element itself should be
            returned instead of the list. Defaults to ``True``.
          pagination_strategy: A Class implementing the :class:`PaginationStrategy`
            ``Protocol``. If ``None``, then automatic pagination is disabled.
            Defaults to :class:`LegacyStrategy`.

        Returns:
          An iterator over the ``FieldPath`` object(s)' data pages
        """

        def f(fpath: FieldPath, blob: dict[str, Any]) -> dict[str, Any]:
            data = fpath._extract_data(blob)

            if type(data) == list and len(data) == 1 and unwrap:
                return data[0]

            return data

        fpaths = list([fpaths] | traverse | map(FieldPath._auto_select) | traverse)
        for page in self.query_json_iter(fpaths, pagination_strategy):
            data = tuple(fpaths | map(functools.partial(f, blob=page)))

            if len(data) == 1:
                yield data[0]
            else:
                yield data
            data = tuple(fpaths | map(functools.partial(f, blob=page)))

            if len(data) == 1:
                yield data[0]
            else:
                yield data

    def _fetch(self, url: str, blob: dict[str, Any]) -> dict[str, Any]:
        resp = self._client.post(
            url, json=blob, headers=default_header(url) | self.headers
        )
        resp.raise_for_status()

        try:
            raw_data = resp.json()

        except JSONDecodeError:
            raise ServerError(
                f"Server ({url}) did not respond with proper JSON"
                f"\nDid you query a proper GraphQL endpoint?"
                f"\n\n{resp.content}"
            )

        if (data := raw_data.get("data")) is None:
            raise GraphQLError(raw_data.get("errors", "Unknown Error(s) Found"))

        return data

    def __enter__(self):
        self._client.__enter__()
        return self

    def __exit__(self, *args):
        self._client.__exit__(*args)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from pipe import where

from subgrounds.schema import SchemaMeta, TypeMeta, TypeRef
from subgrounds.utils import identity

from .fieldpath import FieldPath, SyntheticField

if TYPE_CHECKING:
    from .subgraph import Subgraph


@dataclass
class Object:
    _subgraph: Subgraph
    _object: TypeMeta.ObjectMeta | TypeMeta.InterfaceMeta

    def __init__(
        self, subgraph: Subgraph, object: TypeMeta.ObjectMeta | TypeMeta.InterfaceMeta
    ) -> None:
        self._subgraph = subgraph
        self._object = object

        for field_ in self._object.fields:
            super().__setattr__(field_.name, self._select(field_.name))

    @property
    def _schema(self) -> SchemaMeta:
        return self._subgraph._schema

    def _select(self, name: str) -> FieldPath:
        """Selects the field from ``self`` with name ``name`` and returns the field
        as a :class:`FieldPath`.

        Args:
          name: The name of the field

        Raises:
          TypeError: _description_

        Returns:
          _description_
        """

        field = self._object.field(name)

        match self._schema.type_of_typeref(field.type_):
            case (
                TypeMeta.ObjectMeta()
                | TypeMeta.InterfaceMeta()
                | TypeMeta.EnumMeta()
                | TypeMeta.ScalarMeta() as type_
            ):
                return FieldPath(
                    self._subgraph,
                    TypeRef.Named(name=self._object.name, kind="OBJECT"),
                    field.type_,
                    [(None, field)],
                )

            case TypeMeta.T as type_:
                raise TypeError(
                    f"Object: Unexpected type {type_.name}"
                    f" when selection {name} on {self}"
                )

        assert False  # Suppress mypy missing return statement warning

    def _add_field(self, name: str, fpath: FieldPath) -> None:
        sfield = SyntheticField(identity, fpath._type, fpath)
        self._subgraph._add_synthetic_field(self._object, name, sfield)

    def _add_sfield(self, name: str, sfield: SyntheticField) -> None:
        # TODO: Add check to make sure obj has the deps of sfield
        # obj_fields = [field.name for field in obj.object_.fields]
        # sfield_deps = [fpath.leaf.name for fpath in sfield.deps]
        # for dep in sfield_deps:
        #   if dep not in obj_fields:
        #     raise Exception(
        #       f'SyntheticField {obj.object_.name}.{name}: {obj.object_.name}'
        #        ' does not have the field {dep}'
        #     )

        def f(
            obj_: TypeMeta.ObjectMeta | TypeMeta.InterfaceMeta, path: list[str]
        ) -> None:
            match path:
                case [field_name, *rest]:
                    try:
                        field: TypeMeta.FieldMeta = next(
                            obj_.fields | where(lambda field: field.name == field_name)
                        )

                        f(self._schema.type_map[field.type_.name], rest)
                    except StopIteration:
                        raise Exception(
                            f"SyntheticField {self._object.name}.{name}: {obj_.name}"
                            f" does not have the field {field_name}"
                        )
                case []:
                    return

        for fpath in sfield._deps:
            f(self._object, fpath._name_path())

        self._subgraph._add_synthetic_field(self._object, name, sfield)

    # ================================================================
    # Overloaded magic functions
    # ================================================================
    # Field selection
    def __getattribute__(self, __name: str) -> Any:
        # Small hack to get code completion to work while allowing updates to FieldPath
        # (i.e.: setting arguments)
        try:
            match super().__getattribute__(__name):
                case FieldPath() | SyntheticField() | None:
                    return self._select(__name)
                case value:
                    return value
        except AttributeError:
            return self._select(__name)

    def __setattr__(
        self, __name: str, __value: SyntheticField | FieldPath | Any
    ) -> None:
        match __value:
            case SyntheticField() as sfield:
                self._add_sfield(__name, sfield)
                super().__setattr__(__name, self._select(__name))
            case FieldPath() as fpath:
                self._add_field(__name, fpath)
                super().__setattr__(__name, self._select(__name))
            case _:
                super().__setattr__(__name, __value)



""" Schema data structure module

This module contains various data structures in the form of dataclasses that
are used to represent GraphQL schemas in Subgrounds using an AST-like approach.
"""

from __future__ import annotations

import warnings
from typing import Annotated, Literal

from pipe import map, where
from pydantic.v1 import BaseModel as PydanticBaseModel
from pydantic.v1 import Field, root_validator

from .errors import SchemaError

warnings.simplefilter("default")


class BaseModel(PydanticBaseModel):
    class Config:
        allow_population_by_field_name = True


class TypeRef:
    class T(BaseModel):
        """Base class of all types of type references."""

        @property
        def name(self) -> str:
            raise NotImplementedError

        @property
        def is_list(self) -> bool:
            return False

        @property
        def is_non_null(self) -> bool:
            return False

    class Named(T):
        name_: str = Field(alias="name")
        kind: str

        @property
        def name(self) -> str:
            return self.name_

    class NonNull(T):
        inner: TypeRef_T = Field(alias="ofType")
        kind: Literal["NON_NULL"] = "NON_NULL"

        @property
        def name(self) -> str:
            return self.inner.name

        @property
        def is_list(self) -> bool:
            return self.inner.is_list

        @property
        def is_non_null(self) -> bool:
            return True

    class List(T):
        inner: TypeRef_T = Field(alias="ofType")
        kind: Literal["LIST"] = "LIST"

        @property
        def name(self) -> str:
            return self.inner.name

        @property
        def is_list(self) -> bool:
            return True

        @property
        def is_non_null(self) -> bool:
            return self.inner.is_non_null

    @staticmethod
    def non_null(name: str, kind: str = "SCALAR") -> TypeRef_T:
        return TypeRef.NonNull(inner=TypeRef.Named(name=name, kind=kind))

    @staticmethod
    def non_null_list(name: str, kind: str = "SCALAR") -> TypeRef_T:
        return TypeRef.NonNull(
            inner=TypeRef.List(
                inner=TypeRef.NonNull(inner=TypeRef.Named(name=name, kind=kind))
            )
        )

    @staticmethod
    def root_type_name(type_: TypeRef.T) -> str:
        return type_.name

    @staticmethod
    def is_non_null(type_: TypeRef.T) -> bool:
        return type_.is_non_null

    @staticmethod
    def is_list(type_: TypeRef.T) -> bool:
        return type_.is_list

    @staticmethod
    def graphql(type_: TypeRef.T) -> str:
        match type_:
            case TypeRef.Named(name=name):
                return name
            case TypeRef.NonNull(inner=t):
                return f"{TypeRef.graphql(t)}!"
            case TypeRef.List(inner=t):
                return f"[{TypeRef.graphql(t)}]"

        assert False


TypeRef_ListOrNonNull = Annotated[
    TypeRef.List | TypeRef.NonNull, Field(discriminator="kind")
]

TypeRef_T = TypeRef_ListOrNonNull | TypeRef.Named


class TypeMeta:
    class T(BaseModel):
        """Base class of all GraphQL schema types."""

        name: str
        description: str | None = ""

        @property
        def is_object(self) -> bool:
            return False

    class ArgumentMeta(T):
        """Class representing a field argument definition."""

        type_: TypeRef_T = Field(alias="type")
        default_value: str | None = Field(alias="defaultValue")

    class FieldMeta(T):
        """Class representing an object field definition."""

        arguments: list[TypeMeta.ArgumentMeta] = Field(alias="args")
        type_: TypeRef_T = Field(alias="type")

        def has_arg(self, argname: str) -> bool:
            try:
                next(self.arguments | where(lambda arg: arg.name == argname))
                return True
            except StopIteration:
                return False

        def type_of_arg(self: TypeMeta.FieldMeta, argname: str) -> TypeRef.T:
            try:
                return next(
                    self.arguments
                    | where(lambda argmeta: argmeta.name == argname)
                    | map(lambda arg: arg.type_)
                )
            except StopIteration:
                raise SchemaError(
                    f"TypeMeta.FieldMeta.type_of_arg: no argument named {argname}"
                    f" for field {self.name}"
                )

    class ScalarMeta(T):
        kind: Literal["SCALAR"] = "SCALAR"
        """ Class representing an scalar definition."""

    class ObjectMeta(T):
        """Class representing an object definition."""

        kind: Literal["OBJECT"] = "OBJECT"
        fields: list[TypeMeta.FieldMeta]
        interfaces_: list[dict] = Field(alias="interfaces", default_factory=list)

        # interfaces: list[str] = Field(default_factory=list)

        @property
        def interfaces(self) -> list[str]:
            return list(self.interfaces_ | map(lambda intf: intf["name"]))

        @property
        def is_object(self) -> bool:
            return True

        def field(self, fname: str) -> TypeMeta.FieldMeta:
            """Returns the field definition of object :attr:`self` with name
             :attr:`fname`, if any.

            Args:
                fname: The name of the desired field definition

            Raises:
                KeyError: If no field named :attr:`fname` is defined for object
                  :attr:`self`.

            Returns:
                The field definition
            """

            try:
                return next(self.fields | where(lambda fmeta: fmeta.name == fname))
            except StopIteration:
                raise KeyError(
                    f"TypeMeta.ObjectMeta.field: no field named {fname} for"
                    f" interface {self.name}"
                )

        def type_of_field(self, fname: str) -> TypeRef.T:
            """Returns the type reference of the field of object :attr:`self`
             with name :attr:`fname`, if any.

            Args:
                fname: The name of the desired field type

            Raises:
                KeyError: If no field named :attr:`fname` is defined for object
                 :attr:`self`.

            Returns:
                The field type reference
            """

            try:
                return next(
                    self.fields
                    | where(lambda fmeta: fmeta.name == fname)
                    | map(lambda fmeta: fmeta.type_)
                )
            except StopIteration:
                raise KeyError(
                    f"TypeMeta.ObjectMeta.type_of_field: no field named {fname}"
                    f" for object {self.name}"
                )

    class EnumValueMeta(T):
        """Class representing an enum value definition."""

        pass

    class EnumMeta(T):
        """Class representing an enum definition."""

        kind: Literal["ENUM"] = "ENUM"
        values: list[TypeMeta.EnumValueMeta] = Field(alias="enumValues")

    class InterfaceMeta(T):
        """Class representing an interface definition."""

        kind: Literal["INTERFACE"] = "INTERFACE"
        fields: list[TypeMeta.FieldMeta]

        @property
        def is_object(self) -> bool:
            return False

        def field(self, fname: str) -> TypeMeta.FieldMeta:
            """Returns the field definition of interface `self` with name `fname`,
             if any.

            Args:
                self: The interface type
                fname: The name of the desired field definition

            Raises:
                KeyError: If no field named :attr:`fname` is defined for interface
                  :attr:`self`.

            Returns:
                The field definition
            """

            try:
                return next(self.fields | where(lambda fmeta: fmeta.name == fname))
            except StopIteration:
                raise KeyError(
                    f"TypeMeta.InterfaceMeta.field: no field named {fname}"
                    f" for interface {self.name}"
                )

    class UnionMeta(T):
        """Class representing an union definition."""

        kind: Literal["UNION"] = "UNION"
        types: list[str] = Field(alias="possibleTypes")

    class InputObjectMeta(T):
        """Class representing an input object definition."""

        kind: Literal["INPUT_OBJECT"] = "INPUT_OBJECT"
        input_fields: list[TypeMeta.ArgumentMeta] = Field(alias="inputFields")

        def type_of_input_field(self, fname: str) -> TypeRef.T:
            """Returns the type reference of the input field named `fname` in the
            input object `self`, if any.

            Args:
                fname: The name of the input field

            Raises:
                KeyError: If `fname` is not an input field of input object `self`

            Returns:
                The type reference for input field `fname`
            """

            try:
                return next(
                    self.input_fields
                    | where(lambda infield: infield.name == fname)
                    | map(lambda infield: infield.type_)
                )
            except StopIteration:
                raise KeyError(
                    f"TypeMeta.InputObjectMeta.type_of_input_field: no input field"
                    f" named {fname} for input object {self.name}"
                )


TypeMeta_T = Annotated[
    TypeMeta.ScalarMeta | TypeMeta.ObjectMeta
    # | TypeMeta.EnumValueMeta
    | TypeMeta.EnumMeta
    | TypeMeta.InterfaceMeta
    | TypeMeta.UnionMeta
    | TypeMeta.InputObjectMeta,
    Field(discriminator="kind"),
]

# Pydantic stuff
TypeRef.Named.update_forward_refs()
TypeRef.List.update_forward_refs()
TypeRef.NonNull.update_forward_refs()

TypeMeta.ArgumentMeta.update_forward_refs()
TypeMeta.FieldMeta.update_forward_refs()
TypeMeta.ScalarMeta.update_forward_refs()
TypeMeta.ObjectMeta.update_forward_refs()
TypeMeta.EnumValueMeta.update_forward_refs()
TypeMeta.EnumMeta.update_forward_refs()
TypeMeta.InterfaceMeta.update_forward_refs()
TypeMeta.UnionMeta.update_forward_refs()
TypeMeta.InputObjectMeta.update_forward_refs()


class SchemaMeta(BaseModel):
    """Class representing a GraphQL schema.

    Contains all type definitions."""

    query_type_: dict[str, str] = Field(alias="queryType")
    types: list[TypeMeta_T]
    type_map: dict[str, TypeMeta_T] = Field(default_factory=dict)
    mutation_type_: dict[str, str] | None = Field(alias="mutationType", default=None)
    subscription_type_: dict[str, str] | None = Field(
        alias="subscriptionType", default=None
    )

    @property
    def query_type(self) -> str:
        return self.query_type_["name"]

    @property
    def mutation_type(self) -> str | None:
        if self.mutation_type_:
            return self.mutation_type_["name"]

        return None

    @property
    def subscription_type(self) -> str | None:
        if self.subscription_type_:
            return self.subscription_type_["name"]

        return None

    # @validator("type_map", "types", always=True, pre=True)
    @root_validator(allow_reuse=True)
    def type_map_generator(cls, values):
        if "types" in values and len(values["types"]) > 0:
            values["type_map"] = {type_.name: type_ for type_ in values["types"]}

        return values

    def type_of_typeref(self, typeref: TypeRef.T) -> TypeMeta_T:
        """Returns the type information of the type reference `typeref`

        Args:
            typeref: The type reference pointing to the type of interest.

        Raises:
            KeyError: If the type reference refers to a non-existant type

        Returns:
            Type information
        """

        tname = TypeRef.root_type_name(typeref)

        try:
            return self.type_map[tname]
        except KeyError:
            raise KeyError(
                f"SchemaMeta.type_of_typeref: No type named {typeref.name} in schema!"
            )

    def type_of(
        self: SchemaMeta, tmeta: TypeMeta.ArgumentMeta | TypeMeta.FieldMeta
    ) -> TypeMeta_T:
        """Returns the argument or field definition's underlying type"""

        return self.type_of_typeref(tmeta.type_)

    def type_of_input_object_meta(
        self, tmeta: TypeMeta.InputObjectMeta, args: list[str]
    ) -> TypeRef.T:
        """Recursively finds the nested type"""

        if len(args) < 1:
            raise Exception("type_of_input_object_meta: TODO")

        type_ref = tmeta.type_of_input_field(args.pop(0))
        match self.type_of_typeref(type_ref):
            case TypeMeta.InputObjectMeta() as tmeta:
                return self.type_of_input_object_meta(tmeta, args)

            case _:
                return type_ref



from __future__ import annotations

from dataclasses import replace
from functools import partial
from typing import TYPE_CHECKING, Any, Callable

from pipe import map, traverse

from subgrounds.errors import TransformError
from subgrounds.query import (
    DataRequest,
    DataResponse,
    Document,
    DocumentResponse,
    Query,
    Selection,
)
from subgrounds.schema import TypeMeta, TypeRef
from subgrounds.utils import flatten

from .base import DocumentTransform, RequestTransform
from .utils import select_data

if TYPE_CHECKING:
    from subgrounds.subgraph import Subgraph


class TypeTransform(DocumentTransform):
    """Transform to be applied to scalar fields on a per-type basis.

    Attributes:
      type_: Type indicating which scalar values (i.e.: values of that type) should be
        transformed using the function ``f``
      f: Function to be applied to scalar values of type ``type_`` in the response data.
    """

    type_: TypeRef.T
    f: Callable[[Any], Any]

    def __init__(self, type_: TypeRef.T, f: Callable[[Any], Any]) -> None:
        self.type_ = type_
        self.f = f
        super().__init__()

    def transform_document(self, doc: Document) -> Document:
        return doc

    def transform_response(
        self, doc: Document, resp: DocumentResponse
    ) -> DocumentResponse:
        def transform(select: Selection, data: dict[str, Any]) -> None:
            # TODO: Handle NonNull and List more graciously...
            #   (i.e.: without using `TypeRef.root_type_name``)
            match (select, data):
                # Type matches
                case (
                    Selection(
                        TypeMeta.FieldMeta(name=name, type_=ftype), None, _, [] | None
                    )
                    | Selection(
                        TypeMeta.FieldMeta(type_=ftype), str() as name, _, [] | None
                    ),
                    dict() as data,
                ) if TypeRef.root_type_name(self.type_) == TypeRef.root_type_name(
                    ftype
                ):
                    match data[name]:
                        case list() as values:
                            data[name] = list(
                                values
                                | map(
                                    lambda value: self.f(value)
                                    if value is not None
                                    else None
                                )
                            )
                        case None:
                            data[name] = None
                        case _ as value:
                            data[name] = self.f(value)

                case (Selection(_, _, _, [] | None), dict()):
                    pass

                case (
                    Selection(TypeMeta.FieldMeta(name=name), None, _, inner_select)
                    | Selection(TypeMeta.FieldMeta(), str() as name, _, inner_select),
                    dict() as data,
                ):
                    match data[name]:
                        case list() as elts:
                            for elt in elts:
                                for select in inner_select:
                                    transform(select, elt)
                        case dict() as elt:
                            for select in inner_select:
                                transform(select, elt)
                        case None:
                            return None
                        case _:
                            raise TransformError(
                                f"transform_data_type: data for selection {select} is"
                                f" neither list or dict {data[name]}"
                            )

                case (select, data):
                    raise TransformError(
                        f"transform_data_type:"
                        f" invalid selection {select} for data {data}"
                    )

        for select in doc.query.selection:
            transform(select, resp.data)

        return resp


class LocalSyntheticField(DocumentTransform):
    """Transform used to implement synthetic fields on GraphQL objects that only
    depend on values accessible from that object.

    :meth:`transform_document` replaces the field :attr:`fmeta` by the argument
    fields selections :attr:`args` if the synthetic field :attr:`fmeta` is present
    in the document.

    :meth:`transform_response` applied :attr:`f` to the fields corresponding to
    the argument selections :attr:`args` and places the result in the response.

    Attributes:
      subgraph: The subgraph to which the synthetic field's object belongs.
      fmeta: The synthetic field.
      type_: The object for which the synthetic field is defined.
      f: The function to be applied to the argument fields.
      default: The default value of the synthetic field used in case of exceptions
        (e.g.: division by zero).
      args: The selections of the fields used as arguments to compute the
        synthetic field.
    """

    subgraph: Subgraph
    fmeta: TypeMeta.FieldMeta
    type_: TypeMeta.ObjectMeta | TypeMeta.InterfaceMeta
    f: Callable
    default: Any
    args: list[Selection]

    def __init__(
        self,
        subgraph: Subgraph,
        fmeta: TypeMeta.FieldMeta,
        type_: TypeMeta.ObjectMeta | TypeMeta.InterfaceMeta,
        f: Callable,
        default: Any,
        args: list[Selection],
    ) -> None:
        self.subgraph = subgraph
        self.fmeta = fmeta
        self.type_ = type_
        self.f = f
        self.default = default
        self.args = args

    def transform_document(self, doc: Document) -> Document:
        def transform(select: Selection) -> Selection | list[Selection]:
            match select:
                # case Selection(TypeMeta.FieldMeta(name) as fmeta, _, _, [] | None)
                #  if name == self.fmeta.name and fmeta.type_.name == self.type_.name:
                case Selection(
                    TypeMeta.FieldMeta(name=name), _, _, [] | None
                ) if name == self.fmeta.name:
                    return Selection.merge(self.args)
                case Selection(_, _, _, [] | None):
                    return [select]
                case Selection(
                    TypeMeta.FieldMeta(name=name) as select_fmeta,
                    alias,
                    args,
                    inner_select,
                ):
                    new_inner_select = list(inner_select | map(transform) | traverse)
                    return Selection(select_fmeta, alias, args, new_inner_select)
                case _:
                    raise TransformError(
                        f"transform_document: unhandled selection {select}"
                    )

        def transform_on_type(select: Selection) -> Selection:
            match select:
                case Selection(
                    TypeMeta.FieldMeta(type_=type_) as select_fmeta,
                    alias,
                    args,
                    inner_select,
                ) if type_.name == self.type_.name:
                    new_inner_select = Selection.merge(
                        list(inner_select | map(transform) | traverse)
                    )
                    return Selection(select_fmeta, alias, args, new_inner_select)

                case Selection(fmeta, alias, args, inner_select):
                    return Selection(
                        fmeta, alias, args, list(inner_select | map(transform_on_type))
                    )

        if self.subgraph._url == doc.url:
            return Document.transform(
                doc,
                query_f=lambda query: Query.transform(
                    query, selection_f=transform_on_type
                ),
            )
        else:
            return doc

    def transform_response(
        self, doc: Document, resp: DocumentResponse
    ) -> DocumentResponse:
        def transform(select: Selection, data: dict) -> None:
            match (select, data):
                case (
                    Selection(TypeMeta.FieldMeta(name=name), None, _, [] | None)
                    | Selection(TypeMeta.FieldMeta(), name, _, [] | None),
                    dict() as data,
                ) if name == self.fmeta.name and name not in data:
                    # Case where the selection selects a the synthetic field of
                    #  the current transform that is not in the data blob
                    #  and there are no inner selections

                    # Try to grab the arguments to the synthetic field
                    #  transform in the data blob
                    arg_values = flatten(
                        list(self.args | map(partial(select_data, data=data)))
                    )

                    try:
                        data[name] = self.f(*arg_values)
                    except Exception:
                        data[name] = self.default

                case (
                    Selection(TypeMeta.FieldMeta(name=name), None, _, [] | None)
                    | Selection(TypeMeta.FieldMeta(), name, _, [] | None),
                    dict() as data,
                ) if name not in data:
                    # Case where the selection selects a regular field but it is not in
                    #  the data blob (caused by None value at higher selection)
                    data[name] = None

                case (
                    Selection(TypeMeta.FieldMeta(name=name), None, _, [] | None)
                    | Selection(TypeMeta.FieldMeta(), name, _, [] | None),
                    dict() as data,
                ):
                    # Case where the selection selects a regular field
                    #  and there are no inner selections (nothing to do)
                    pass

                case (
                    Selection(TypeMeta.FieldMeta(name=name), None, _, inner_select)
                    | Selection(TypeMeta.FieldMeta(), name, _, inner_select),
                    dict() as data,
                ) if name in data:
                    match data[name]:
                        case list() as elts:
                            for elt in elts:
                                for select in inner_select:
                                    transform(select, elt)

                        case dict() as elt:
                            for select in inner_select:
                                transform(select, elt)

                        case None:
                            data[name] = {}
                            for select in inner_select:
                                transform(select, data[name])

                        case _:
                            raise TransformError(
                                f"transform_response: data for selection {select} is"
                                f" neither list or dict {data[name]}"
                            )

                case (select, data):
                    raise TransformError(
                        f"transform_response: invalid selection {select}"
                        f" for data {data}"
                    )

        def transform_on_type(
            select: Selection, data: list[Any] | dict[str, Any]
        ) -> None:
            match select:
                case Selection(TypeMeta.FieldMeta(type_=type_), None, _, _) | Selection(
                    TypeMeta.FieldMeta(type_=type_), _, _, _
                ) if type_.name == self.type_.name:
                    # for select in inner_select:
                    #   transform(select, data[name])
                    match data:
                        case list():
                            for d in data:
                                transform(select, d)
                        case dict():
                            transform(select, data)

                case (
                    Selection(TypeMeta.FieldMeta(name=name), None, _, inner_select)
                    | Selection(_, name, _, inner_select)
                ):
                    match data:
                        case list():
                            for d in data:
                                list(
                                    inner_select
                                    | map(partial(transform_on_type, data=d[name]))
                                )
                        case dict():
                            list(
                                inner_select
                                | map(partial(transform_on_type, data=data[name]))
                            )

        if self.subgraph._url == doc.url:
            for select in doc.query.selection:
                transform_on_type(select, resp.data)

        return resp


class DocumentRequestTransform(RequestTransform):
    """This class serves as a wrapper for :class:`DocumentTransform` for ease of use.

    Internally, we convert all :class:`DocumentTransform`s to
     :class:`DocumentRequestTransform`s due to the relative ease it has on our
     implementation. We merely iterate through the documents within a request and
     response iterating them thru the relevant documents (based upon url).
    """

    def __init__(self, transform: DocumentTransform, url: str):
        self.transform = transform
        self.url = url

    def transform_request(self, req: DataRequest):
        return replace(
            req,
            documents=[
                self.transform.transform_document(doc)
                for doc in req.documents
                if doc.url == self.url
            ],
        )

    def transform_response(self, req: DataRequest, data: DataResponse):
        """

        Assumes 1 doc -> 1 resp, and ordering is correct via the `zip`
        """

        return replace(
            data,
            responses=[
                self.transform.transform_response(doc, resp)
                for doc, resp in zip(req.documents, data.responses)
                if resp.url == self.url  # TODO: doc.url != resp.url, exception
            ],
        )



class SubgroundsError(Exception):
    """The base error for all subgrounds errors"""


class SchemaError(SubgroundsError):
    """Errors related to schema"""


class TransformError(SubgroundsError):
    """Errors related to transforms"""


class ServerError(SubgroundsError):
    """Errors returned by a server"""


class GraphQLError(ServerError):
    """Errors returned by a GraphQL server"""



""" Subgraph module that defines various classes to manipulate requests and
subgraphs.

This module is the glue that connects the lower level modules (i.e.:
:module:`query`, :module:`schema`, :module:`transform`, :module:`pagination`) to
the higher toplevel modules (i.e.: :module:`subgrounds`).
"""

from __future__ import annotations

import logging
import warnings
from dataclasses import dataclass, field

from pipe import map

from subgrounds.schema import SchemaMeta, TypeMeta, TypeRef
from subgrounds.transform import (
    DEFAULT_SUBGRAPH_TRANSFORMS,
    DocumentTransform,
    LocalSyntheticField,
)

from .fieldpath import FieldPath, SyntheticField
from .object import Object

logger = logging.getLogger("subgrounds")
warnings.simplefilter("default")


@dataclass
class Subgraph:
    _url: str
    _schema: SchemaMeta
    _transforms: list[DocumentTransform] = field(default_factory=list)
    _is_subgraph: bool = True

    def __init__(
        self,
        url: str,
        schema: SchemaMeta,
        transforms: list[DocumentTransform] = DEFAULT_SUBGRAPH_TRANSFORMS,
        is_subgraph: bool = True,
    ) -> None:
        self._url = url
        self._schema = schema
        self._transforms = transforms
        self._is_subgraph = is_subgraph

        # Add objects as attributes
        for key, obj in self._schema.type_map.items():
            match obj:
                case TypeMeta.ObjectMeta() | TypeMeta.InterfaceMeta():
                    super().__setattr__(key, Object(self, obj))
                case _:
                    pass

    def _add_synthetic_field(
        self,
        object_: TypeMeta.ObjectMeta | TypeMeta.InterfaceMeta,
        name: str,
        sfield: SyntheticField,
    ) -> None:
        fmeta = TypeMeta.FieldMeta(
            name=name, description="", args=[], type=sfield._type
        )
        object_.fields.append(fmeta)

        sfield_fpath = FieldPath(
            self,
            TypeRef.Named(name=object_.name, kind="OBJECT"),
            sfield._type,
            [(None, fmeta)],
        )
        logger.debug(
            "Subgraph: Adding SyntheticField at FieldPath"
            f" {sfield_fpath._root_type.name}.{sfield_fpath._name_path()}"
        )

        transform = LocalSyntheticField(
            self,
            fmeta,
            object_,
            sfield._f,
            sfield._default,
            list(sfield._deps | map(FieldPath._selection)),
        )

        self._transforms = [transform, *self._transforms]



from typing import Any

import plotly.graph_objects as go
from pipe import map, traverse

from subgrounds import Subgrounds
from subgrounds.query import DataRequest

from .traces import TraceWrapper


class Figure:
    subgrounds: Subgrounds
    traces: list[TraceWrapper]
    req: DataRequest | None
    data: list[dict[str, Any]] | None
    figure: go.Figure

    args: dict[str, Any]

    def __init__(
        self,
        subgrounds: Subgrounds,
        traces: TraceWrapper | list[TraceWrapper],
        **kwargs
    ) -> None:
        self.subgrounds = subgrounds
        self.traces = list([traces] | traverse)

        traces = list(self.traces | map(lambda trace: trace.field_paths) | traverse)

        if len(traces) > 0:
            self.req = self.subgrounds.mk_request(traces)
            self.data = self.subgrounds.execute(self.req)
        else:
            self.req = None
            self.data = None

        self.args = kwargs
        self.refresh()

    def refresh(self) -> None:
        # TODO: Modify this to support x/y in different documents
        self.figure = go.Figure(**self.args)

        if self.req is None:
            return

        self.data = self.subgrounds.execute(self.req)

        for trace in self.traces:
            self.figure.add_trace(trace.mk_trace(self.data))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['subgrounds',
 'subgrounds.client',
 'subgrounds.contrib.dash',
 'subgrounds.contrib.plotly',
 'subgrounds.pagination',
 'subgrounds.subgraph',
 'subgrounds.transform']

package_data = \
{'': ['*'], 'subgrounds': ['contrib/*', 'resources/*']}

install_requires = \
['httpx[http2]>=0.25.1,<1.0.0',
 'pandas>=2.1,<3.0',
 'pipe>=2.0,<3.0',
 'pydantic>=2.0,<3.0',
 'pytest-asyncio']

extras_require = \
{'all': ['dash>=2.3.1,<3.0.0', 'plotly>=5.14.1,<6.0.0'],
 'dash': ['dash>=2.3.1,<3.0.0'],
 'plotly': ['plotly>=5.14.1,<6.0.0']}

setup_kwargs = {
    'name': 'subgrounds-gnosis',
    'version': '1.9.2',
    'description': 'A Pythonic data access layer for applications querying data from The Graph Network.',
    'long_description': '# Subgrounds\n\nFork of https://github.com/0xPlaygrounds/subgrounds with small changes due to pending PRs in the original repository.\n\n<!-- [![GitHub Actions](https://github.com/gnosis/subgrounds/workflows/CI/badge.svg)](https://github.com/gnosis/subgrounds/actions) -->\n[![PyPI](https://img.shields.io/pypi/v/subgrounds.svg)](https://pypi.org/project/subgrounds/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/subgrounds.svg)](https://pypi.org/project/subgrounds/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n[![CI](https://github.com/gnosis/subgrounds/actions/workflows/main.yml/badge.svg)](https://github.com/gnosis/subgrounds/actions/workflows/main.yml)\n<br>\n\n[![Discord](https://img.shields.io/discord/896944341598208070?color=7289DA&label=discord&logo=discord&logoColor=fff)](https://discord.gg/gMSSh5bjvk)\n[![Twitter Follow](https://img.shields.io/badge/Playgrounds-Analytics-31fa1f2Playgrounds0x?color=%231fa1f2&logo=Twitter&logoColor=1fa1f2&style=flat)](https://twitter.com/Playgrounds0x)\n\n[![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/gnosis/subgrounds/blob/main/examples/notebook.ipynb)\n[![Github Codepsaces](https://img.shields.io/badge/Github-Codespaces-24292f.svg?logo=Github)](https://codespaces.new/gnosis/subgrounds-template?quickstart=1)\n\n<!-- start elevator-pitch -->\nAn intuitive Python library for interfacing with subgraphs and GraphQL.\n\n## Features\n- **Simple**: Leverage a Pythonic API to easily build queries and transformations without the need for raw GraphQL manipulation.\n- **Automated**: Automatically handle pagination and schema introspection for effortless data retrieval.\n- **Powerful**: Create sophisticated queries using the `SyntheticFields` transformation system.\n<!-- end elevator-pitch -->\n\n## Resources\n- [**Docs**](http://docs.playgrounds.network/): User guide and API documentation\n- [**Snippets**](https://github.com/gnosis/subgrounds/tree/main/examples): A list of examples showcasing Subgrounds integration with Dash and Plotly\n- [**Examples**](http://docs.playgrounds.network/subgrounds/examples/): An ever growing list of projects created by our community members and team\n- [**Videos**](https://docs.playgrounds.network/subgrounds/videos/): Video workshops on Subgrounds\n\n## Installation\n> Subgrounds **requires** atleast Python 3.10+\n\nSubgrounds is available on PyPi. To install it, run the following:<br>\n`pip install subgrounds-gnosis`.\n\nSubgrounds also comes bundled with extra modules that may require extra libraries. You can get all functionality of `subgrounds` via the following:<br>\n`pip install subgrounds-gnosis[all]`.\n\n## Simple example\n<!-- start simple-example -->\n```python\n>>> from subgrounds import Subgrounds\n\n>>> sg = Subgrounds()\n\n>>> # Load\n>>> aave_v3 = sg.load_subgraph("https://api.thegraph.com/subgraphs/name/messari/aave-v3-ethereum")\n\n>>> # Construct the query\n>>> largest_markets = aave_v3.Query.markets(\n...     orderBy=aave_v3.Market.totalValueLockedUSD,\n...     orderDirection=\'desc\',\n...     first=5,\n... )\n\n>>> # Return query to a dataframe\n>>> sg.query_df([\n...     largest_markets.name,\n...     largest_markets.totalValueLockedUSD,\n... ])\n          markets_name  markets_totalValueLockedUSD\n0 Aave Ethereum wstETH                 1.915090e+09\n1   Aave Ethereum WETH                 1.047417e+09\n2   Aave Ethereum WBTC                 6.049655e+08\n3   Aave Ethereum USDC                 4.482461e+08\n4   Aave Ethereum USDT                 4.280103e+08\n```\n<!-- end simple-example -->\n\n\n## About Us\nPlaygrounds Analytics is a data solutions company providing data infrastructures and services for data teams, analysts, and engineers. Check us out [here](https://playgrounds.network/) to learn more!\n\n\n## Acknowledgments\nThis software project would not be possible without the support of The Graph Foundation. You can learn more about The Graph and its mission [here](https://thegraph.com/).\n',
    'author': 'cvauclair',
    'author_email': 'cvauclair@playgrounds.network',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/gnosis/subgrounds',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



""" Toplevel Subgrounds module

This module implements the toplevel API that most developers will be using when
querying The Graph with Subgrounds.
"""

import warnings

from .client import Subgrounds

__all__ = ["Subgrounds"]

warnings.warn(
    "Importing from `subgrounds.subgrounds` is deprecated."
    " Use `subgrounds` instead.\n"
    "Will be removed in a future version.",
    DeprecationWarning,
)



from __future__ import annotations

import logging
import warnings
from dataclasses import dataclass
from enum import Enum, auto
from functools import reduce
from typing import TYPE_CHECKING, Any

from subgrounds.schema import TypeMeta
from subgrounds.utils import merge

if TYPE_CHECKING:
    from subgrounds.subgraph import FieldPath

logger = logging.getLogger("subgrounds")
warnings.simplefilter("default")


@dataclass
class Filter:
    fields: list[TypeMeta.FieldMeta]
    op: Filter.Operator
    value: Any

    class Operator(Enum):
        EQ = auto()
        NEQ = auto()
        LT = auto()
        LTE = auto()
        GT = auto()
        GTE = auto()

    @classmethod
    def mk_filter(cls, fpath: FieldPath, op: Filter.Operator, value: Any) -> Filter:
        match fpath._path:
            case [(_, TypeMeta.FieldMeta()), *_] as path:  # assert atleast len of 1
                return cls([tup[1] for tup in path], op, value)
            case _:
                raise TypeError(
                    f"Cannot create filter on FieldPath {fpath}: not a native field!"
                )

    @property
    def leaf_name(self):
        leaf = self.fields[-1].name
        match self.op:
            case Filter.Operator.EQ:
                return leaf
            case Filter.Operator.NEQ:
                return f"{leaf}_not"
            case Filter.Operator.LT:
                return f"{leaf}_lt"
            case Filter.Operator.GT:
                return f"{leaf}_gt"
            case Filter.Operator.LTE:
                return f"{leaf}_lte"
            case Filter.Operator.GTE:
                return f"{leaf}_gte"

    @classmethod
    def to_dict(cls, filters: list[Filter]) -> dict[str, Any]:
        """Converts a series of filters into a single dict.

        Each filter represents a single key, value pair where the key is generated based
          on the operation (via :func:`~subgrounds.subgraph.filter.Filter.leaf_name`).
          For nested filters, this base dict is further stacked under sub-objects:

        ```py
        {"outer_": {"outer2_": {"inner_op": "value"}}}
        ```
        """

        return reduce(
            merge,
            (
                reduce(
                    lambda x, y: {f"{y.name}_": x},
                    filter.fields[:-1],
                    {filter.leaf_name: filter.value},
                )
                for filter in filters
            ),
            {},
        )



"""
DEPRECIATED: Use `subgrounds.contrib.dash` instead
"""

import warnings

from subgrounds.contrib.dash import AutoUpdate, DataTable, Graph, Refreshable

__all__ = [
    "Refreshable",
    "Graph",
    "DataTable",
    "AutoUpdate",
]

warnings.warn(
    "Importing from `subgrounds.plotly_wrappers` is deprecated."
    " Use `subgrounds.contrib.plotly` instead.\n"
    "Will be removed in a future version.",
    DeprecationWarning,
)



from .fieldpath import FieldPath, SyntheticField
from .filter import Filter
from .object import Object
from .subgraph import Subgraph

__all__ = [
    "FieldPath",
    "Filter",
    "Object",
    "Subgraph",
    "SyntheticField",
]
